@@ -2,8 +2,41 @@ import argparse
 import sys, os, imp, uuid
 import six
 import collections
+from six.moves.collections_abc import Sized, Iterable, Container, Mapping
 import functools
 import types
+import warnings
+
+__version__ = "2.4.0"
+
+# from https://github.com/python/cpython/blob/main/Lib/_collections_abc.py
+# missing from python < 3.6
+def _check_methods(C, *methods):
+    mro = C.__mro__
+    for method in methods:
+        for B in mro:
+            if method in B.__dict__:
+                if B.__dict__[method] is None:
+                    return NotImplemented
+                break
+        else:
+            return NotImplemented
+    return True
+
+class Collection(Sized, Iterable, Container):
+    __slots__ = ()
+    @classmethod
+    def __subclasshook__(cls, C):
+        if cls is Collection:
+            return _check_methods(C,  "__len__", "__iter__", "__contains__")
+        return NotImplemented
+
+# from numpy
+class VisibleDeprecationWarning(UserWarning):
+    pass
+
+def deprecation(message):
+    warnings.warn(message, VisibleDeprecationWarning, stacklevel=2)
 
 # from https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-subobjects-chained-properties/31174427#31174427
 def _rgetattr(obj, attr, *args):
@@ -16,18 +49,101 @@ class MagiConfigError(Exception):
     pass
 
 class MagiConfig(argparse.Namespace):
-    def write(self, filename, config_obj):
+    def write(self, filename, config_obj, attr_imports=None, class_imports=None, attr_reprs=None, class_reprs=None, strict=False):
         if len(config_obj)==0:
             raise MagiConfigError("config_obj must be specified")
+
+        # defaults
+        if attr_imports is None: attr_imports = {}
+        if class_imports is None: class_imports = {}
+        if attr_reprs is None: attr_reprs = {}
+        if class_reprs is None: class_reprs = {}
+
+        # get lines to write
+        default_imports = ["from magiconfig import MagiConfig"]
+        imports, lines = self._write(config_obj, attr_imports, class_imports, attr_reprs, class_reprs, strict)
+
         # write namespace into file
         with open(filename,'w') as outfile:
-            lines = []
-            prepend = ""
-            # create a magiconfig
-            lines.extend(["from magiconfig import MagiConfig","",config_obj+" = MagiConfig()"])
-            prepend = config_obj + "."
-            lines.extend([prepend+str(attr)+" = "+repr(val) for attr,val in sorted(six.iteritems(vars(self)))])
-            outfile.write('\n'.join(lines))
+            outfile.write('\n'.join(default_imports+sorted(list(imports))+[""]+lines))
+
+    def _write(self, config_obj, attr_imports, class_imports, attr_reprs, class_reprs, strict):
+        imports = set()
+        # create a magiconfig
+        lines = [config_obj+" = MagiConfig()"]
+        prepend = config_obj + "."
+        for attr,val in sorted(six.iteritems(vars(self))):
+            valclass = val.__class__
+            # recurse for nested configs
+            if valclass==self.__class__:
+                imports_tmp, lines_tmp = val._write(prepend+attr, attr_imports, class_imports, attr_reprs, class_reprs, strict)
+                imports.update(imports_tmp)
+                lines.extend(lines_tmp)
+            else:
+                # precedence: attr-specific > class-specific > default
+                imports.update(self._get_imports(attr, val, attr_imports, class_imports))
+                repr_fn = attr_reprs.get(attr, class_reprs.get(valclass, repr))
+                repr_val = repr_fn(val)
+                # detect cases where repr() doesn't work as desired
+                if strict:
+                    try:
+                        repr_worked = eval(repr_val)
+                    # catch cases where eval(repr) completely fails
+                    except:
+                        repr_worked = False
+                    if not repr_worked:
+                        error_lines = [
+                            "Inconsistency between original value and repr for {}:".format(prepend+attr),
+                            val,
+                            repr_val,
+                        ]
+                        raise MagiConfigError('\n'.join(error_lines))
+                lines.append(prepend+str(attr)+" = "+repr_val)
+        return imports, lines
+
+    # recursively check imports (avoiding infinite recursion in self-referential case)
+    def _get_imports(self, attr, val, attr_imports, class_imports, checked=None):
+        valclass = val.__class__
+        if checked is None: checked = set()
+        imports = set()
+
+        def default_import(val):
+            valclass = val.__class__
+            if valclass.__module__=='__builtin__':
+                if valclass.__name__=='module':
+                    return "import {}".format(val.__name__)
+            else:
+                return "from {} import {}".format(valclass.__module__,valclass.__name__)
+            return None
+
+        if id(val) not in checked:
+            # precedence: attr-specific > class-specific > defaults
+            attr_imports_actual = {} if attr is None else attr_imports
+            import_fn = attr_imports_actual.get(attr, class_imports.get(valclass, default_import))
+            imports_to_add = import_fn(val)
+            if imports_to_add is not None:
+                imports.add(imports_to_add)
+
+            checked.add(id(val))
+
+            # check collection entries
+            coll = None
+            if isinstance(val, Mapping):
+                coll = six.iteritems(val)
+            elif isinstance(val, Collection):
+                coll = val
+            elif isinstance(val, Container):
+                try:
+                    coll = vars(val)
+                except:
+                    # todo: solution for this case?
+                    pass
+            if coll is not None:
+                for subval in coll:
+                    # will be applied to keys and values separately for dicts, i.e. attr-specific only use for MagiConfigs
+                    imports.update(self._get_imports(None, subval, attr_imports, class_imports, checked))
+
+        return imports
 
     # to merge with another config
     def join(self, other_config, prefer_other=False):
@@ -115,10 +231,9 @@ class ArgumentParser(argparse.ArgumentParser):
         self._config_only_help = kwargs.pop("config_only_help", True)
         # must be defined before base class constructor is called
         self._dests_actions = collections.defaultdict(list)
-        self._config_only = {}
+        self._config_only = collections.OrderedDict()
         argparse.ArgumentParser.__init__(self, *args, **kwargs)
         self._config_actions = None
-        self._config_only_required = set()
 
         # initialize config args from options
         self._init_config()
@@ -312,9 +427,9 @@ class ArgumentParser(argparse.ArgumentParser):
                 tmp = val
                 # check type if uniquely provided
                 if len(self._dests_actions[attr])==1 or len(set([action.type for action in self._dests_actions[attr]]))==1:
-                    # use _get_values() rather than _get_value() to handle nargs cases
+                    # use _get_values() rather than _get_value() to handle nargs cases; also enforces choices if any
                     tmp_action = self._dests_actions[attr][0]
-                    # argparse does not apply type checks to default args
+                    # argparse does not apply type or choice checks to default args
                     if tmp==tmp_action.default:
                         pass
                     # nargs=0 is usually _StoreTrueAction or _StoreFalseAction:
@@ -328,8 +443,6 @@ class ArgumentParser(argparse.ArgumentParser):
                         # _get_values() expects a list
                         if not isinstance(tmp,list): tmp = [tmp]
                         tmp = self._get_values(tmp_action,tmp)
-                    # enforce choices if any
-                    self._check_value(tmp_action, tmp)
                 setattr(namespace,attr,tmp)
                 possible_required_actions.extend(self._dests_actions[attr])
             else:
@@ -338,7 +451,7 @@ class ArgumentParser(argparse.ArgumentParser):
         self._required = self._suppress_required(possible_required_actions)
 
         # check missing required config-only args
-        config_only_missing = self._config_only_required - set([attr for attr in flat_vars])
+        config_only_missing = set([dest for dest,action in six.iteritems(self._config_only) if action.required]) - set([attr for attr in flat_vars])
         if len(config_only_missing)>0:
             raise MagiConfigError("Imported config missing required attributes: "+','.join(sorted(list(config_only_missing))))
 
@@ -369,42 +482,108 @@ class ArgumentParser(argparse.ArgumentParser):
         self._init_config()
 
     # write namespace into file using config_obj
-    def write_config(self, namespace, filename, obj=None):
+    def write_config(self, namespace, filename, obj=None, attr_imports=None, class_imports=None, attr_reprs=None, class_reprs=None, strict=False):
         if obj is None:
             if self.config_options is not None: obj = self.config_options.obj
             else: obj = "config"
-        namespace.write(filename,obj)
+        namespace.write(filename, obj, attr_imports, class_imports, attr_reprs, class_reprs, strict)
 
     # add config-only arguments
     # args: no default value, not required
     # kwargs: default value OR required (value=None)
     def add_config_only(self, *args, **kwargs):
+        deprecation("ArgumentParser.add_config_only() is deprecated and will be removed in magiconfig 3.0.0; please switch to ArgumentParser.add_config_argument().")
+
         # check for existing dests
         existing_dests = [arg for arg in list(args) + list(kwargs) if arg in self._dests_actions]
-        if len(existing_dests)>0: raise MagiConfigError("the following dests are already used by regular (not config-only) args: "+', '.join(existing_dests))
+        if len(existing_dests)>0: raise MagiConfigError("the following dests are already used by regular (not config-only) arguments: "+', '.join(existing_dests))
 
-        self._config_only.update(kwargs)
-        self._config_only.update({key:None for key in args})
-
-        # find any args that have changed required status
-        self._config_only_required = (self._config_only_required | set([key for key in kwargs if kwargs[key] is None])) - set([key for key in kwargs if kwargs[key] is not None])
+        for dest in args:
+            self.add_config_argument(dest)
+        for dest,default in six.iteritems(kwargs):
+            if default is None:
+                self.add_config_argument(dest, required=True)
+            else:
+                self.add_config_argument(dest, default=default)
 
     # remove config-only argument
     def remove_config_only(self, arg):
-        self._config_only.pop(arg)
-        if arg in self._config_only_required: self._config_only_required.remove(arg)
+        deprecation("ArgumentParser.remove_config_only() is deprecated and will be removed in magiconfig 3.0.0; please switch to ArgumentParser.remove_config_argument().")
+
+        self.remove_config_argument(arg)
+
+    def _get_config_only_kwargs(self, arg, **kwargs):
+        return dict(kwargs, dest=arg, option_strings=[])
+
+    # add a config-only argument
+    # available properties: default, type, choices, required, help
+    # mostly based on argparse add_argument()
+    def add_config_argument(self, arg, **kwargs):
+        kwargs = self._get_config_only_kwargs(arg, **kwargs)
+
+        # if no default was supplied, use the parser-level default
+        if 'default' not in kwargs:
+            dest = kwargs['dest']
+            if dest in self._defaults:
+                kwargs['default'] = self._defaults[dest]
+            elif self.argument_default is not None:
+                kwargs['default'] = self.argument_default
+
+        # if not required, treat as optional (needed for some argparse handling, e.g. help w/ no default specified)
+        if not kwargs.get('required',False):
+            kwargs['nargs'] = argparse.OPTIONAL
+
+        # create the (dummy) action object, and add it to the parser
+        action_class = self._pop_action_class(kwargs)
+        if not callable(action_class):
+            raise ValueError('unknown action "%s"' % (action_class,))
+        action = action_class(**kwargs)
+
+        # raise an error if the action type is not callable
+        type_func = self._registry_get('type', action.type, action.type)
+        if not callable(type_func):
+            raise ValueError('%r is not callable' % (type_func,))
+
+        if type_func is FileType:
+            raise ValueError('%r is a FileType class object, instance of it'
+                             ' must be passed' % (type_func,))
+
+        action = self._add_config_only_action(action)
+        if action.help is None:
+            if action.required: action.help = "(required)"
+            elif action.default is not None: action.help = " " # must be non-None string to activate default formatting
+            #else: action.help = " "
+
+        return action
 
     # keep map of dest:action(s)
     _add_action_orig = argparse.ArgumentParser._add_action
 
     def _add_action(self, action):
         # check against config-only
-        if action.dest in self._config_only: raise argparse.ArgumentError(action, "dest {} already specified as config-only".format(action.dest))
+        if action.dest in self._config_only: raise argparse.ArgumentError(action, "dest {} already specified as config-only argument".format(action.dest))
 
         action = self._add_action_orig(action)
         self._dests_actions[action.dest].append(action)
         return action
 
+    def _add_config_only_action(self, action):
+        # check for existing dests
+        if action.dest in self._config_only: raise argparse.ArgumentError(action, "conflicting config-only dest: {}".format(action.dest))
+        if action.dest in self._dests_actions: raise argparse.ArgumentError(action, "dest {} already specified as regular (not config-only) argument".format(action.dest))
+
+        # add to actions list
+        self._config_only[action.dest] = action
+        action.container = self
+
+        return action
+
+    def remove_config_argument(self, arg):
+        if arg in self._config_only:
+            self._config_only.pop(arg)
+        else:
+            self.error("attempt to remove unrecognized config-only argument: {}".format(arg))
+
     _remove_action = _remove_action_all
 
     # allow removing single argument
@@ -453,25 +632,8 @@ class ArgumentParser(argparse.ArgumentParser):
         # config-only args
         if len(self._config_only)>0 and self._config_only_help:
             formatter.start_section("config-only arguments")
-            # convert to fake actions to make use of existing HelpFormatter methods
-            config_only_actions = []
-            for arg in sorted(self._config_only):
-                kwarg = dict(
-                    option_strings = [],
-                    dest = arg,
-                )
-                if arg in self._config_only_required:
-                    kwarg.update(
-                        required = True,
-                        help = "(required)",
-                    )
-                elif self._config_only[arg] is not None:
-                    kwarg.update(
-                        default = self._config_only[arg],
-                        nargs = argparse.OPTIONAL,
-                        help = " ", # must be non-None string to activate default formatting
-                    )
-                config_only_actions.append(argparse._StoreAction(**kwarg))
+            # get list of (dummy) actions
+            config_only_actions = [action for dest,action in six.iteritems(self._config_only)]
             formatter.add_arguments(config_only_actions)
             formatter.end_section()
 
